# TBNode objects and reading from file.
#
# This module attempts to give a clean interface to interacting with penn treebank style 
# data in perl.
# 
# Author: scott cotton
# Date: 20030707
#

package TBNode;
use strict; # make perl have some semblence of sanity

#
# constructor, inheritable style
#
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    $self->{LABEL} = shift;
    $self->{DATA} = shift;
    $self->{PARENT} = undef;
    $self->{CHILDREN} = [];
    $self->{NEXTSIBLING} = undef;
    $self->{PREVSIBLING} = undef;
    $self->{TERMNO} = undef;
    $self->{HEIGHT} = undef;
    return $self;
}

#
# get/set the label associated with the node, this is the 
# whole label with function tags and all
#
sub rawLabel {
    my $self = shift;
    if (@_) { $self->{LABEL} = shift; }
    return $self->{LABEL};
}

#
# return the primary label of this node
#
sub label {
    my $self = shift;
    my @parts = split('-', $self->rawLabel());
    return $parts[0];
}


#
# get the function tags associated with the node, including NONE, and
# any numeric trace identifiers
#
sub ftags {
    my $self = shift;
    my @tags = split('-', $self->rawLabel());
    return @tags[1 .. $#tags];
}


#
# get/set the data associated with the node
#
sub data {
    my $self = shift;
    if (@_) { $self->{DATA} = shift; }
    return $self->{DATA};
}


#
# get/set the "terminal number" of the node
# this is the index of the least terminal (empty constituents included) 
# which is dominated by this node.
#
sub termno {
    my $self = shift;
    if (@_) { $self->{TERMNO} = shift; }
    return $self->{TERMNO};
}


#
# get/set the height associated with this node.
#
sub height {
    my $self = shift;
    if (@_) { $self->{HEIGHT} = shift; }
    return $self->{HEIGHT};
}


#
# get/set the parent of this node
# returns undef if last sibling
sub parent {
    my $self = shift;
    if (@_) { $self->{PARENT} = shift; }
    return $self->{PARENT};
}

#
# get/set the right sibling of this node
# returns undef if last sibling
#
sub nextSibling {
    my $self = shift;
    if (@_) { $self->{NEXTSIBLING} = shift; }
    return $self->{NEXTSIBLING};
}

#
# get/set the left sibling of this node
# returns undef if this node is leftmost
#
sub prevSibling {
    my $self = shift;
    if (@_) { $self->{PREVSIBLING} = shift; }
    return $self->{PREVSIBLING};
}

#
# get/set the children of this node
#
sub children {
    my $self = shift;
    if (@_) { @{$self->{CHILDREN}} = @_; }
    return @{$self->{CHILDREN}};
}

#
# the first child of this node if it is not a terminal,
# otherwise return undef
#
sub firstChild {
    my $self = shift;
    my @kids = $self->children();
    if ($#kids >= 0) {
        return $kids[0];
    } else {
        return undef;
    }
}

#
# get the root node of this syntax tree
#
sub root {
    my $self = shift;
    my $node = $self;
    while(defined $node->parent()) {
        $node = $node->parent();
    }
    return $node;
}

#
# return the highest index of the list of this nodes children
# eg, a node with a single child would return 0,
# a node with no children would return -1
#
sub nchildren {
    my $self = shift;
    my @foo = $self->children();
    return $#foo;
}

#
# add a child to this node, linking next/prev-Sibling and parent-child
# relations accordingly.
#
sub addChild {
    my $self = shift;
    my $new_child = shift;
    $new_child->parent($self);
    my @kids = $self->children();
    push(@kids, $new_child);
    $self->children(@kids);
    if ($#kids > 0) {
        my $last_child = $kids[$#kids - 1];
        $last_child->nextSibling($new_child);
        $new_child->prevSibling($last_child);
    }
}

#
# print out some stuff that looks vaguely like a treebank file
#
sub show {
    my $self = shift;
    print "(", $self->rawLabel(), " ";
    if ($self->nchildren() >= 0) {
        for my $child ($self->children()) {
            $child->show();
        }
        print ")";
    } else {
        print $self->data(), ")";
    }
}


#
# used by nextSentence below to set the terminal number and height of
# internal and root nodes.  not for external use.
#
sub percolate {
    my $self = shift;
    my $lastnode = $self;
    my $node = $self->parent();
    while(defined($node)) {
        last if (defined($node->height()));
        $node->height($lastnode->height() + 1);
        $node->termno($lastnode->termno());
        $lastnode = $node;
        $node = $node->parent();
    }
}


#
# given a predicate and optionall a list ref,
# populate the list with all the nodes either dominated by
# this node or this node which satisfy the predicate.
# return the list.
#
sub find {
    my $self = shift;
    my $predicate = shift;
    my $res = shift || []; 
    if (&$predicate($self)) {
        push(@{$res}, $self);
    }
    for my $child ($self->children()) {
        $child->find($predicate, $res);
    }
    return @{$res};
}

#
# iterate code over this node and all its descendents
#
sub iter {
    my $self = shift;
    my $code = shift;
    &$code($self);
    for my $child ($self->children()) {
        $child->iter($code);
    }
}


#
# return the trace index associated with this node,
# or undef if there is no such thing.
#
sub traceId {
    my $self = shift;
    my @ftags = $self->ftags();
    for my $tag (@ftags) {
        return $tag if $tag =~ /^\d+$/;
    }
    if ($self->data() =~ /^\*.*-(\d+).*/) { return $1; }
    return undef;
}

#
# find all the nodes in this sentence with the same trace id as this node
# 
sub traceIdEquivs {
    my $self = shift;
    my $tid = $self->traceId();
    my $predref = sub { 
        my $node = shift;
        return ($node->traceId() == $tid);
    };
    if (defined ($tid)) { return $self->root()->find($predref); }
    return ($self);
}

#
# return true iff this node has no children
#
sub isTerminal {
    my $self = shift;
    return $self->nchildren() == -1;
}


#
# return 1 iff this node is "a trace", ie a node without
# children, without a primary label, with a "NONE" function
# tag label.  By convention, nodes with "NONE" in the function
# tags never have primary labels.
#
sub isTrace {
    my $self = shift;
    if (! $self->isTerminal() ) { return 0; }
    for my $tag ($self->ftags()) {
        if ($tag == "NONE") {
                return 1;
        }
    }
    return 0;
}


#
# return true iff this node has exactly one
# child which is a trace
#
sub isEmptyNonTerm {
    my $self = shift;
    return $self->nchildren() == 0 && $self->firstChild()->isTrace();
}

#
# follow a numeric trace.
# eg, if we have a tree (A (B-2 b) (C-1 (-NONE- *-2)) (D ...) (E (-NONE- *T*-1)))
# and we take the node (-NONE- *T*-1) as $self, then we return the 
# array ((-NONE- *T*-1), (C-1 (-NONE- *-2)), (-NONE- *-2), (B-2 b))
#
sub followNumeric {
    my $self = shift;
    my %done = ($self => 1);
    my $found = 1;
    my @current = ($self);
    my @res=($self);
    while($found) {
        $found = 0;
        my @tidnodes = ();
        foreach my $c (@current) { push(@tidnodes, $c->traceIdEquivs()); }
        @current = ();
        foreach my $tn (@tidnodes) {
            if ($tn->isEmptyNonTerm() && !defined($done{$tn->firstChild()})) {
                push(@current, $tn->firstChild());
                $found = 1;
            }
            if (!defined($done{$tn})) {
                $done{$tn} = 1;
                push(@res, $tn);
            }
        }
    }
    return @res;
}


sub rawdata {
    my $self = shift;
    my $isterm = sub { my $node = shift; return $node->height() == 0; };
    my @terms = $self->find($isterm);
    my $result = "";
    for(my $i=0; $i<$#terms; $i++) {
        $result .= $terms[$i]->data() . " ";
    }
    $result .= $terms[$#terms]->data();
}


#
# a "static" method which finds the next root node from a file handle
#
sub nextSentence {
    my $fh = shift;
    my @node_stack = ();
    my $buf;
    my $readres;
    my $char;
    my $in_sent=0;
    my $nlabel;
    my $root;
    my $terminal=0;
    while (($readres = read($fh, $char, 1)) != 0) {
        die "unable to read from $fh! $!\n" unless (defined $readres);
        if ($char eq "(") {
            $nlabel = "";
            $in_sent = 1;
            my $firstwhite = 1;
            while(($readres = read($fh, $char, 1)) != 0) {
                die "unable to read from $fh! $!\n" unless (defined $readres);
                if ($firstwhite) {
                    if ($char =~ /\s/) { next; }
                    elsif ($char eq "(") {
                        my $node = TBNode->new("TOP");
                        if ($#node_stack >= 0) {
                            my $nparent = $node_stack[$#node_stack];
                            $nparent->addChild($node);
                        } else {
                            $root = $node;
                        }
                        push (@node_stack, $node);
                        $firstwhite=1;
                        $nlabel="";
                        next;
                    }
                    $nlabel .= $char;
                    $firstwhite=0;
                    next;
                }
                last if ($char =~ /\s/);
                $nlabel .= $char;
            }
            my $node = TBNode->new($nlabel);
            if ($#node_stack >= 0) {
                my $nparent = $node_stack[$#node_stack];
                $nparent->addChild($node);
            } else {
                $root = $node;
            }
            push(@node_stack, $node);
        } elsif ($in_sent && $char eq ")") {
            pop @node_stack;
            if ($#node_stack < 0) { $in_sent=0; last; }
        } elsif ($in_sent && $char !~ /\s/) {
            my $contents = $char;
            while(($readres = read($fh, $char, 1)) != 0) {
                die "unable to read from $fh! $!\n" unless (defined $readres);
                next if ($char =~ /\s/);
                if ($char eq ")") {
                    my $node = pop @node_stack;
                    if ($#node_stack < 0) { $in_sent = 0; }
                    $node->data($contents);
                    $node->termno($terminal);
                    $node->height(0);
                    $node->percolate();
                    $terminal++;
                    last;
                } 
                $contents .= $char;
            }
        } else { next; }
    }
    if ($#node_stack != -1) {
        die "mismatched parens. $#node_stack\n";
    }
    return $root;
}

=pod

=head1 TBNode Module

This module provides a clean perl interface to penn treebank data.  It follows
closely both the Ocaml and Java interfaces.  The module provides two pieces of
functionality

=over 4

=item 1.
    Parsing of treebank files.

=item 2.
    An OOP interface to nodes in the treebank.

=back

B<Author:> Scott Cotton
B<Date:> 20030707

=head2 Parsing of treebank files

The interface for parsing treebank files is as follows:
Note you can use any file handle, not just STDIN.

    use TBNode;
    while (defined($tbn = TBNode::nextSentence(STDIN))) {
        # do stuff with a tbnode object $tbn ...
    }

The parser for these parses recognizes I<any> opening 
parenthesis as an indicator that a sentence is coming.
Additionally, the parser doesn't like nodes without
two parts, like the string "(foo)" somewhere in the text.
Finally, for sentences with a blank outer layer of
parenthesis, a dummy label "TOP" is inserted.


=head2 Using TBNode objects

In this sections, all variables named with "tbn" as a prefix, eg C<$tbn>, are
assumed to be treebank nodes.

=over 4

=item Tree-like interface:

    # return the parent node, or undef if root
    $tbn_parent = $tbn->parent(); 

    # return the first child of $tbn, or undef if $tbn is a terminal
    $tbn_first = $tbn->firstChild(); 

    # return the right sibling, or undef if already rightmost
    $tbn = $tbn->nextSibling();

    # return left sibling, or undef if already leftmost
    $tbn = $tbn->prevSibling();

    # return all the children
    @tbnodes = $tbn->children();

    # return the root of the tree
    $root = $tbn->root();

=item Label and Data interfaces:

    # return the primary label of a treebank, eg "NP"
    $tbn->label();

    # return the raw label, with function tags and all
    $tbn->rawLabel();

    # return the function tags, including trace identifiers, etc
    @ftags = $tbn->ftags();

    # return the terminal number associated with the node,
    $termno = $tbn->termno();

    # return the height of the node from the leftmost
    # dominated terminal, with terminals ranked with height C<0>.
    $height = $tbn->height();

    # return the data associated with the node, (undef if
    # $tbn has children).  EG for the node (NNP John), return "John".
    $data = $tbn->data();

    # return the raw data in the nodes, concatenate all the terminals
    # data together, joined by spaces:
    $data = $tbn->raw_data();

=item Traces, empty constituents:

    # return the traceId of this node, or undef if there is one
    $tid = $tbn->traceId();

    # return an array of all the nodes in $tbn's sentence with
    # trace identifier equal to that of $tbn (if it is defined).
    # NOTE it is always the case that $tbn is a member of the
    # result, even if $tbn's trace identifier is undefined.
    @nodes = $tbn->traceIdEquivs();

    # follow a numeric trace anchored at $tbn through the tree.
    @tracechain = $tbn->followNumeric();

    # return true iff $tbn is a node without a main label, with
    # "NONE" function tag, and without any children.
    $tbn->isTrace();

    # return true iff $tbn has exactly one child which is a trace
    $tbn->isEmptyNonTerm();

=item Iterating, Filtering.

    #
    # return list of tbnode objects satisfying both "has a label matching NP"
    # and which are among the nodes dominated by $tbn or $tbn itself.
    # search goes in depth first, left to right order.
    #
    # (NOTE trailing semi-colon on next line of code!)
    #
    my $code = sub { my $node = shift; return $node->label() =~ /NP/; };
    @NPs = $tbn->find($code);

    # iterate a formatter over all the tbnodes
    # in depth first, left to right order
    my $fmter = sub { my $node = shift; print $node->termno(), "\n"; };
    $tbn->iter($fmter);

=back

=head2 NOTES:

This module arose from the need to document and make generally usable
for dumb people like myself code interfacing "the treebank".  If you
happen to find it helpful, please consider being kind to your fellow
coder by

=over 4

=item  B<1. documenting> code!

=item B<2. hiding> hacky useless information.

=back
    
=cut

# to make perl happy
1;
